-- Advanced Functional Programming course 2020
-- Chalmers TDA342 / GU DIT260
--
-- 2020-02-17 Guest lecture by Andreas Abel
--
-- Introduction to Agda
--
-- File 1: The Curry-Howard Isomorphism

{-# OPTIONS --allow-unsolved-metas #-}

module Prelude where

-- Natural numbers as our first example of
-- an inductive data type.

data ℕ : Set where  -- \ b N
  zero : ℕ
  suc  : ℕ → ℕ

-- To use decimal notation for numerals, like
-- 2 instead of (suc (suc zero)), connect it
-- to the built-in natural numbers.

{-# BUILTIN NATURAL ℕ #-}

-- Lists are a parameterized inductive data type.

data List (A : Set) : Set where
  []  : List A
  _∷_ : (a : A) (as : List A) → List A   -- \ : :

infixr 6 _∷_

-- foo : List ℕ → Set
-- foo [] = {!!}
-- foo (a ∷ y) = {!!}

-- C-c C-l   load
-- C-c C-c   case split
-- C-c C-,   goal and context
-- C-c C-.   goal and context and expression
-- C-c C-SPC give
-- C-c C-a   auto

-- Disjoint sum type.

data _⊎_ (S T : Set) : Set where  -- \uplus
  inl : S → S ⊎ T
  inr : T → S ⊎ T

infixr 4 _⊎_  -- A ⊎ B ⊎ C = A ⊎ (B ⊎ C)

case' : (S T U : Set) → S ⊎ T → (f : S → U) → (g : T → U) → U
case' S T U (inl x) f g = f x
case' S T U (inr x) f g = g x

-- case : (S T U : Set) → S ⊎ T → (f : S → U) → (g : T → U) → U
-- case S T U (inl x) = λ f g → f x
-- case S T U (inr x) = λ f g → g x

-- case : (S T U : Set) → S ⊎ T → (f : S → U) → (g : T → U) → U
-- case S T U = λ{ (inl x) f g → f x
--               ; (inr x) f g → g x
--               }

case : {S T U : Set} → S ⊎ T → (f : S → U) → (g : T → U) → U
case {S} {T} {U} (inl x) f g = f x
case {U = V} (inr x) f g = g x


-- The empty sum is the type with 0 alternatives,
-- which is the empty type.
-- By the Curry-Howard-Isomorphism,
-- which views a proposition as the set/type of its proofs,
-- it is also the absurd proposition.

data False : Set where

⊥-elim : False → {A : Set} → A
⊥-elim ()

-- C-c C-SPC give
-- C-c C-, show hypotheses and goal
-- C-c C-. show hypotheses and infers type

-- Tuple types

-- The generic record type with two fields
-- where the type of the second depends on the value of the first
-- is called Sigma-type (or dependent sum), in analogy to
--
--   Σ ℕ A =  Σ   A(n) = A(0) + A(1) + ...
--           n ∈ ℕ

record Σ (A : Set) (B : A → Set) : Set where  -- \GS  \Sigma
  constructor _,_
  field  fst : A
         snd : B fst
open Σ

infixr 5 _,_

data Unit : Set where
  unit : Unit

IsZero : ℕ → Set
IsZero zero    = Unit
IsZero (suc n) = False

foo : Σ ℕ IsZero
foo = record { fst = 0; snd = unit }

-- C-c C-r refine

-- The non-dependent version is the ordinary Cartesian product.

_×_ : (S T : Set) → Set
S × T = Σ S (λ _ → T)

infixr 5 _×_

-- The record type with no fields has exactly one inhabitant
-- namely the empty tuple record{}
-- By Curry-Howard, it corresponds to Truth, as
-- no evidence is needed to construct this proposition.

record True : Set where

test : True
test = record{}

-- Example: distributivity  A ∧ (B ∨ C) → (A ∧ B) ∨ (A ∧ C)

dist : ∀{A B C : Set} → A × (B ⊎ C) → (A × B) ⊎ (A × C)
dist (a , inl b) = inl (a , b)
dist (a , inr c) = inr (a , c)

-- Relations

-- Type-theoretically, the type of relations 𝓟(A×A) is
--
--   A × A → Prop
--
-- which is
--
--   A × A → Set
--
-- by the Curry-Howard-Isomorphism
-- and
--
--   A → A → Set
--
-- by currying.

Rel : (A : Set) → Set₁
Rel A = A → A → Set

-- Less-or-equal on natural numbers

_≤_ : Rel ℕ
zero  ≤ y     = True
suc x ≤ zero  = False
suc x ≤ suc y = x ≤ y

-- C-c C-l load
-- C-c C-c case split
-- C-c C-, show goal and assumptions
-- C-c C-. show goal and assumptions and current term
-- C-c C-SPC give

-- -}
-- -}
